/* Creator:  Wade Tregaskis
   Created:  Mon 7th, May 2002
   Last Modified:  Tue 9th, May 2006

   Description:  An menu-based CLI app which allows for the manipulation of a
	   database of song information - e.g. title, track, artist, etc.
         Also has functions for converting between seconds and formatted track
      time.  Primarily uses the former internally.

   Note: Rewrote from scratch using updated playlist 3 source as base.  Too many
      changes were made to playlist 3 after the initial playlist 4 branching,
      making it simpler to copy the few new things from playlist 4 back into
      playlist 3 than vice versa.
   Morale: Do things right the first time, and be prepared to be stuck with what
      you create.

   Note: There is really that much here that it should be implemented in both
      multiple files and as classes instead of structs and functions.  But,
      alas, the use of these is disallowed. :( */

//  Copyright (c) 2006, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cctype>
#include <cstring>
#include <cmath>
using namespace std;


#define LINE_BUFFER_SIZE 80
#define FILENAME_BUFFER_SIZE 255


char *defaultTrackListName = "Default";
char *defaultTrackListFile = "songdb.dat";


struct Track;
struct TrackListItem;
struct TrackList;


/* The system used here deserves some brief documentation.  Each Track is unique
   in memory and stored only once.  When added to a list, a TrackListItem is
   created which points to the track.  This TrackListItem forms a linked list
   controlled by TrackList.  This system, while slightly beefier than
   integrating Track and TrackListItem together, is far more elegant and
   flexible.  It allows, for example, a temporary list to be created when
   searching for all occurances of a track, which is a great aid.  Were support
   for multiple lists to be integrated in a future version, this would make
   such a thing very simple to code.
     Because memory management is of concern, even for this little app, each
   Track contains a claim parameter.  This is incremented every time the track
   is referenced by a TrackListItem, and decremented whenever that TrackListItem
   is destroyed.  This prevents a track occuring in multiple lists from being
   erroneously released when it is removed from just one list.  It adds a little
   bit of coding overhead, but seeing as there's only a few places which create
   or destroy TrackListItem's, this is not significant nor ongoing. */

struct Track {
    int lengthInSeconds;
    char *title, *artist;
    bool warning; // Means Track may contain invalid data due to input error
    int claims;
};

struct TrackListItem {
    TrackListItem *previous, *next, *original;
    TrackList *owner;
    Track *target;
};

struct TrackList {
    TrackListItem *first, *last;
    int numberOfTracks;
    char *name, *openedFrom;
};


char displayMenu(void);
void displayRepeat(ostream& out, char ch, int times);

void uiSecondsToFormattedTime(void);
void uiFormattedTimeToSeconds(void);
void uiDisplaySongDatabase(TrackList *list);
void uiAddSongToDatabase(TrackList *list);
void uiDeleteSongFromDatabase(TrackList *list);
void uiSortSongDatabase(TrackList *list);

void removeOutsideWhitespace(char **data);
void outputDoubleDigit(ostream& out, int number);
void outputFixedLength(ostream& out, const char *data, int maxLength,
                       bool left);
void outputFormattedTime(ostream& out, int hours, int minutes, int seconds,
                         bool negative);

void parseFormattedTimeFromString(const char *data, int& hours, int& minutes,
                                  int& seconds, bool& negative,
                                  bool& possibleError);
int parseSecondsFromString(const char data[], bool& possibleError);
int formattedTimeToSeconds(int hours, int minutes, int seconds, bool negative);
void secondsToFormattedTime(int totalTime, int& hours, int& minutes,
                            int& seconds, bool& negative);
void verifyFormattedTime(int& hours, int& minutes, int& seconds);

TrackList* newTrackList(char *listName, char *sourceFile);
TrackList* newTrackListFromFile(char *listName, char *fileName);
void deleteTrackList(TrackList *target);

void deleteTrackListItem(TrackListItem *item);

Track* newTrack(int length);
Track* newTrack(char *name);
Track* newTrack(char *name, char *artist, int length, bool warning);
void deleteTrack(Track *target);

TrackListItem* findTrackByLength(TrackList *list, const int length);
TrackListItem* findTrackByTitle(TrackList *list, const char *name);
TrackListItem* findTrackByArtist(TrackList *list, const char *name);
TrackListItem* findTrackByIndex(TrackList *list, int val);
TrackListItem* findNextTrackByLength(const int length, TrackListItem *start);
TrackListItem* findNextTrackByTitle(const char *name, TrackListItem *start);
TrackListItem* findNextTrackByArtist(const char *name, TrackListItem *start);
TrackList* findAllTracksByLength(TrackList *list, const int length);
TrackList* findAllTracksByTitle(TrackList *list, const char *name);
TrackList* findAllTracksByArtist(TrackList *list, const char *name);

void addTrackToList(TrackList *list, Track *entry);
void copyTrackToList(TrackList *list, TrackListItem *entry);
bool removeTrackFromList(TrackListItem *entry);
void deleteAllTracksFromList(TrackList *list);
bool readTrackListFromFile(TrackList *list, const char *fileName);
bool saveTrackListToFile(TrackList *list, const char *fileName);
bool saveDatabaseAs(TrackList *list);

char* copyOfString(char *data);
int caseInsensitiveStringCompare(const char *a, const char *b);
void safeGetline(istream& in, char *data, int maxLength);


/* main(void)
     Handles the main menu.  Also loads the default track list, and saves it
     again when quitting. */

int main(void) {
    char choice;
    TrackList *defaultList;

    defaultList = newTrackListFromFile(copyOfString(defaultTrackListName),
                                       copyOfString(defaultTrackListFile));
    
    if (!defaultList) {
        cout << "\nError: Unable to open default song list - it may be missing "
             << "or in use.\nIf you continue, a new [empty] list will be "
             << "created.\n\nDo you wish to continue (y/n)? ";
        cin >> choice;

        if ((choice != 'Y') && (choice != 'y')) {
            return 0;
        }

        defaultList = newTrackList(copyOfString(defaultTrackListName),
                                   copyOfString(defaultTrackListFile));
        choice = '\0';
        cin.ignore(10000, '\n');
    }

    do {
        switch (choice = displayMenu()) {
        case '1':
            uiSecondsToFormattedTime();
            break;
        case '2':
            uiFormattedTimeToSeconds();
            break;
        case '3':
            uiDisplaySongDatabase(defaultList);
            break;
        case '4':
            uiAddSongToDatabase(defaultList);
            break;
        case '5':
            uiDeleteSongFromDatabase(defaultList);
            break;
        case '6':
            uiSortSongDatabase(defaultList);
            break;
        case '7':
            if (!saveDatabaseAs(defaultList)) {
                choice = '0';
            }
            
            break;
        default:
            if (iscntrl(choice)) { // control chars don't output well
                cout << "You entered an invalid selection.  Please try again.";
            } else {
                cout << "You entered an invalid selection (\"" << choice
                     << "\").  Please try again.";
            }
        }
    } while (choice != '7');

    deleteTrackList(defaultList);
    
    return 0;
}

/* displayMenu(void)
     Draws the menu to screen, handling all formatting and whitespace.  Also
     handles user input, returning the selection made.  [Note: It does not
     check that this selection is valid] */

char displayMenu(void) {
    char choice;

    cout << endl << endl;
    displayRepeat(cout, '*', 40); // Top border
    cout << "\n* Playlist Generator                   *"; // While it may be
    cout << "\n* ------------------                   *"; // 'smarter' to use
    cout << "\n*                                      *"; // displayRepeat()
    cout << "\n* 1: Convert seconds to formatted time *"; // for the whitespace,
    cout << "\n* 2: Convert formatted time to seconds *"; // it's quicker &
    cout << "\n*                                      *"; // clearer to just
    cout << "\n* 3: Display Song Database             *"; // format the menu
    cout << "\n* 4: Add Song to Database              *"; // in this fashion.
    cout << "\n* 5: Delete Song from Database         *";
    cout << "\n* 6: Sort Song Database                *";
    cout << "\n*                                      *";
    cout << "\n* 7: Exit Program                      *\n";
    displayRepeat(cout, '*', 40); // Bottom border

    cout << "\n\nPlease select an option (1-7):  ";
    cin >> choice;

    cin.ignore(10000, '\n');
    cout << endl; // Better once here than 6x in each individual procedure

    return choice;
}

/* uiSecondsToFormattedTime(void)
     Asks the user for an integer representing a number of seconds, and converts
     this input to a formatted time scale of format hh:mm:ss.  Outputs the
     result. */

void uiSecondsToFormattedTime(void) {
    char *input;
    int timeInSeconds;
    int hours, minutes, seconds;
    bool isNegative, error;

    cout << "Enter the number of seconds: ";

    input = new char[LINE_BUFFER_SIZE];
    safeGetline(cin, input, LINE_BUFFER_SIZE);

    if (strlen(input)) {
        timeInSeconds = parseSecondsFromString(input, error);
        secondsToFormattedTime(timeInSeconds, hours, minutes, seconds,
                               isNegative);

        if (error) {
            if (timeInSeconds != 0) {
                cout << "Your input was not exactly correct.\nIf you meant to "
                     << "enter \"" << timeInSeconds << "\", that is "
                     << "represented as ";
                outputFormattedTime(cout, hours, minutes, seconds, isNegative);
            } else {
                cout << "Your input didn't resolve to a value.  Try again.";
            }
        } else {
            cout << timeInSeconds << " seconds is represented as ";
            outputFormattedTime(cout, hours, minutes, seconds, isNegative);
        }
    } else {
        cout << "Treating lack of input as a cancel - returning to main menu.";
    }

    delete[] input;
}

/* uiFormattedTimeToSeconds(void)
     The reverse function to uiSecondsToFormattedTime [obviously] - inputs a
     time value of hh:mm:ss format, and converts it to a simple number of
     seconds. */

void uiFormattedTimeToSeconds(void) {
    char *input;
    int hours = 0, minutes = 0, seconds = 0, totalTime = 0;
    bool isNegative, inputMayBeInvalid;
    
    cout << "Enter the formatted (H:M:S) time: ";

    input = new char[LINE_BUFFER_SIZE];
    safeGetline(cin, input, LINE_BUFFER_SIZE);

    if (strlen(input)) {
        parseFormattedTimeFromString(input, hours, minutes, seconds, isNegative,
                                     inputMayBeInvalid);

        totalTime = formattedTimeToSeconds(hours, minutes, seconds, isNegative);
        verifyFormattedTime(hours, minutes, seconds);

        if (inputMayBeInvalid) {
            if (totalTime) {
                cout << "Your input was not exactly correct.\nIf you meant to "
                     << "enter \"";
                outputFormattedTime(cout, hours, minutes, seconds, isNegative);
                cout << "\", that is the representation of " << totalTime
                     << " seconds";
            } else {
                cout << "Your input did not resolve to a value.  Try again.";
            }
        } else {
            outputFormattedTime(cout, hours, minutes, seconds, isNegative);
            cout << " is the representation of " << totalTime << " seconds";
        }
    } else {
        cout << "Treating lack of input as a cancel - returning to main menu.";
    }

    delete[] input;
}

/* uiDisplaySongDatabase(TrackList *list)
     Presents a list of all entries in 'list' to the user.  Note: It does not
     sort the database or interpret the results in any other way than needed to
     display them. */

void uiDisplaySongDatabase(TrackList *list) {
    TrackListItem *currentItem;
    Track *current;
    int hours, minutes, seconds, totalLength = 0;
    bool isNegative, needToExplainWarning = false;

    if (!(list->numberOfTracks)) {
        cout << "There are no songs currently in the list \"" << list->name
             << "\".";
    } else {
        currentItem = list->first;

        cout << "The list \"" << list->name << "\" contains the following "
             << "songs:\n";

        while (currentItem) {
            current = currentItem->target;
            
            totalLength += current->lengthInSeconds;

            secondsToFormattedTime(current->lengthInSeconds, hours, minutes,
                                   seconds, isNegative);

            cout << endl;
            outputFixedLength(cout, current->artist, 30, true);
            cout << "- ";
            outputFixedLength(cout, current->title, 30, true);
            cout << "- ";
            outputFormattedTime(cout, hours, minutes, seconds, isNegative);

            if (current->warning) {
                cout << " *";
                needToExplainWarning = true;
            }

            currentItem = currentItem->next;
        }

        cout << endl << endl;
        displayRepeat(cout, ' ', 57);
        cout << "Total: ";
        secondsToFormattedTime(totalLength, hours, minutes, seconds,
                               isNegative);
        outputFormattedTime(cout, hours, minutes, seconds, isNegative);

        if (needToExplainWarning) {
            cout << "\n\n[Note: Songs marked with a star may have incorrect "
                 << "lengths due to possible\n       errors in the database "
                 << "file]";
        }
    }
}

/* uiAddSongToDatabase(TrackList *list)
     Inputs the title, artist and length of a new track, appending it to
     'list'. */

void uiAddSongToDatabase(TrackList *list) {
    char *artist, *title, *length, check;
    int hours = 0, minutes = 0, seconds = 0, totalTime = 0;
    bool isNegative, error;

    cout << "Adding new song to the list \"" << list->name << "\".\n\n";
    
    artist = new char[LINE_BUFFER_SIZE];
    cout << "Enter Artist Name: ";
    safeGetline(cin, artist, LINE_BUFFER_SIZE);
    removeOutsideWhitespace(&artist);

    if (strlen(artist)) {
        title = new char[LINE_BUFFER_SIZE];
        cout << "Enter Song Title: ";
        safeGetline(cin, title, LINE_BUFFER_SIZE);
        removeOutsideWhitespace(&title);

        if (strlen(title)) {
            length = new char[LINE_BUFFER_SIZE];

            do {
                cout << "Enter Song Length (H:M:S): ";
                safeGetline(cin, length, LINE_BUFFER_SIZE);

                parseFormattedTimeFromString(length, hours, minutes, seconds,
                                             isNegative, error);
                verifyFormattedTime(hours, minutes, seconds);

                if (error) {
                    cout << "\nYour input was not exactly correct.  Did you "
                         << "perhaps mean to enter ";
                    outputFormattedTime(cout, hours, minutes, seconds,
                                        isNegative);
                    cout << " (y/n)?  ";

                    cin >> check;
                    cin.ignore(10000, '\n');

                    error = !(check == 'y' || check == 'Y');
                }
            } while (error);

            if (strlen(length)) {
                totalTime = formattedTimeToSeconds(hours, minutes, seconds,
                                                   isNegative);

                addTrackToList(list, newTrack(title, artist, totalTime, error));

                cout << "\n\"" << title << "\" by \"" << artist
                     << "\" (length ";
                outputFormattedTime(cout, hours, minutes, seconds, isNegative);
                cout << ") added to list \"" << list->name << "\".\nIt now "
                     << "contains " << list->numberOfTracks << " songs.";
            } else {
                cout << "Treating lack of input as a cancel - returning to "
                     << "main menu.";
            }

            delete[] length;
        } else {
            cout << "Treating lack of input as a cancel - returning to main "
                 << "menu.";
        }

        delete[] title;
    } else {
        cout << "Treating lack of input as a cancel - returning to main menu.";
    }

    delete[] artist;
}

/* uiDeleteSongFromDatabase(TrackList *list)
     Allows the user to select an entry from 'list' based on distinguishing
     characteristics (i.e. title and artist) and then removes this entry. */

void uiDeleteSongFromDatabase(TrackList *list) {
    char *artist, *title;
    int hours, minutes, seconds;
    bool isNegative;
    TrackList *result;
    TrackListItem *selectionItem;
    Track *selection;

    cout << "Delete song from list \"" << list->name << "\".\n\n";
    
    artist = new char[LINE_BUFFER_SIZE];
    cout << "Enter Artist Name: ";
    safeGetline(cin, artist, LINE_BUFFER_SIZE);

    if (strlen(artist)) {
        result = findAllTracksByArtist(list, artist);

        if (result->numberOfTracks) {
            selectionItem = result->first;

            cout << "\nFound the following song(s) by \"" << artist << "\":\n";

            while (selectionItem) {
                selection = selectionItem->target;

                cout << endl;
                cout.setf(ios::left);
                cout.width(30);
                cout << selection->title;
                cout << " - ";
                secondsToFormattedTime(selection->lengthInSeconds, hours,
                                       minutes, seconds, isNegative);
                outputFormattedTime(cout, hours, minutes, seconds, isNegative);

                selectionItem = selectionItem->next;
            }

            title = new char[LINE_BUFFER_SIZE];
            cout << "\n\nEnter Song Title: ";
            safeGetline(cin, title, LINE_BUFFER_SIZE);

            if (strlen(title)) {
                selectionItem = findTrackByTitle(result, title);

                if (selectionItem) {
                    cout << "The song \"" << selectionItem->target->title
                         << "\" by \"" << selectionItem->target->artist
                         << "\" has been removed from the list \""
                         << list->name << "\".";

                    removeTrackFromList(selectionItem->original);
                    deleteTrackListItem(selectionItem->original);
                } else {
                    cout << "\nThere is no song by \"" << artist
                         << "\" called \"" << title << "\" in the list \""
                         << list->name << "\".";
                }
            } else {
                cout << "Treating lack of input as a cancel - returning to "
                     << "main menu.";
            }
            
            delete[] title;
        } else {
            cout << "\nThere are no songs by \"" << artist << "\" in the the "
                 << "list \"" << list->name << "\".";
        }

        deleteTrackList(result);
    } else {
        cout << "Treating lack of input as a cancel - returning to main menu.";
    }

    delete[] artist;
}

/* uiSortSongDatabase(TrackList *list)
     Will one day allow the user to sort 'list' in ascending or descending order
     based on any attribute - e.g. title, artist, song length or any other. */

void uiSortSongDatabase(TrackList *list) {
    cout << "Sort song database\nFunction not implemented yet.";
}

/* displayRepeat(ostream& out, char ch, int times)
     Outputs to the stream 'out' a sequence of identical characters - 'char ch'.
     It does so to the tune of 'times'.
       Note: The use of the 'times' parameter instead of a local variable is not
     erroneous. */

void displayRepeat(ostream& out, char ch, int times) {
    while (times > 0) {
        out << ch;
        --times;
    }
}

/* outputDoubleDigit(ostream& out, int number)
     Prints to the output stream 'out' the number specificed, prefixing a 0 when
     needed to ensure the output is always at least two digits. */

void outputDoubleDigit(ostream& out, int number) {
    if (number < 10) {
        out << '0' << number;
    } else {
        out << number;
    }
}

/* outputFixedLength(ostream& out, const char *data, int maxLength, bool left)
     Outputs the cstring 'data' to 'out', establishing an output of fixed
     length, set by 'maxLength', and aligned to left or right, depending on
     'left'.  Most useful for the fact that if 'data' is too long it will be
     truncated and an elipse automagically appended to indicate this. */

void outputFixedLength(ostream& out, const char *data, int maxLength,
                       bool left) {
    char *result;
    int i;

    if (left) {
        cout.setf(ios::left);
    } else {
        cout.setf(ios::right);
    }

    cout.width(maxLength);

    if (strlen(data) > maxLength) { // Need to truncate it
        result = new char[maxLength + 1];
        strncpy(result, data, maxLength - 3);

        for (i = maxLength - 3; i < maxLength; ++i) {
            result[i] = '.';
        }

        result[maxLength] = '\0';
        
        cout << result;

        delete[] result;
    } else { // Fine as it is
        cout << data;
    }
}

/* parseFormattedTimeFromString(const char *data, int& hours, int& minutes,
                                int& seconds, bool& negative,
                                bool& possibleError)
     Takes a pointer to a char array and proceeds to extract track length
     information - of H:M:S form - from it.  Allows for all variations of this
     basic input, such as just M:S or S alone. */

void parseFormattedTimeFromString(const char *data, int& hours, int& minutes,
                                  int& seconds, bool& negative,
                                  bool& possibleError) {
    int value[3] = {0, 0, 0};
    int curValue = 2, placeValue = 1, offset = 0, stopAt = 0;
    bool happy = true;

    if (data[0] == '-') {
        negative = true;
        stopAt = 1;
    } else {
        negative = false;
    }

    possibleError = false;

    for(offset = strlen(data) - 1; (offset >= stopAt) && happy; --offset) {
        if (data[offset] == ':') {
            placeValue = 1;

            if ((--curValue) == -1) {
                happy = false;
            }
        } else {
            if (isdigit(data[offset])) {
                value[curValue] += (data[offset] - 48) * placeValue;
                placeValue *= 10;
            } else {
                possibleError = true; // we encountered something non-parsable
            }
        }
    }

    hours = value[0];
    minutes = value[1];
    seconds = value[2];
}

/* parseSecondsFromString(const char data[], bool& possibleError)
     Basically performs the same functionality as atoi() on 'data'.  What this
     adds is error checking to determine if the original input was perhaps
     invalid.  If so, 'possibleError' is set true.  Returns the value read. */

int parseSecondsFromString(const char data[], bool& possibleError) {
    int offset = 0, length;
    bool dataStarted = false;

    length = strlen(data);
    possibleError = false;

    for(offset = 0; offset < length; ++offset) {
        if (isalpha(data[offset])) {
            if (dataStarted || (data[offset] != '-') && (data[offset] != '+')) {
                possibleError = true;
                dataStarted = true;
            }
        }
    }

    return atoi(data);
}

/* removeOutsideWhitespace(char **data)
     Takes the 'data' and removes any whitespace before or after any text in
     it.  Note that 'data' is a pointer to a pointer to a string (phew!), so
     for a normal *char pointer, you'll need to prefix & to the pointer name
     when passing to this function. */

void removeOutsideWhitespace(char **data) {
    int textStart, textEnd = 0, i, length, newLength;
    char *result, *realData;

    realData = *data;
    length = strlen(realData);

    textStart = length;

    for (i = 0; i < length; ++i) {
        if (!isspace(realData[i])) {
            textStart = i;
            break;
        }
    }

    for (i = length; i > 0; --i) {
        if (!isspace(realData[i - 1])) {
            textEnd = i;
            break;
        }
    }

    newLength = (textEnd - textStart);

    if (newLength > 0) { // malloc doesn't like negative values
        result = new char[newLength + 1];
        
        for (i = 0; i < newLength; ++i) {
            result[i] = realData[textStart + i];
        }

        result[newLength] = '\0';
    } else { // Indicates the original string was pure whitespace
        result = new char[1];
        result[0] = '\0';
    }

    delete[] realData;
    *data = result;
}

/* formattedTimeToSeconds(int hours, int minutes, int seconds, bool negative)
     Converts a series of values for hours, minutes and seconds into a single
     value, representing the total number of seconds.  Returns the result. */

int formattedTimeToSeconds(int hours, int minutes, int seconds, bool negative) {
    if (negative) {
        return -((hours * 3600) + (minutes * 60) + seconds);
    } else {
        return ((hours * 3600) + (minutes * 60) + seconds);
    }
}

/* secondsToFormattedTime(int totalTime, int& hours, int& minutes,
                          int& seconds, bool& negative)
     Converts a value of any number of seconds to separate fields of hours,
     minutes and seconds.  Also preserves the sign value of the input. */

void secondsToFormattedTime(int totalTime, int& hours, int& minutes,
                            int& seconds, bool& negative) {
    int temp;

    negative = (totalTime < 0);
    temp = abs(totalTime);

    hours = temp / 3600; // The round-off due to integer division is desired
    temp %= 3600; // After hours we're left with minutes & seconds
    minutes = temp / 60;
    seconds = temp % 60;
}

/* displayFormattedTime(ostream& out, int hours, int minutes, int seconds,
                        bool negative)
     Very simply outputs to the stream 'out' the hours, minutes and seconds
     fields as a properly displayed formatted time. */

void outputFormattedTime(ostream& out, int hours, int minutes, int seconds,
                         bool negative) {
    if (negative) {
        out << '-';
    }

    outputDoubleDigit(out, hours);
    out << ':';
    outputDoubleDigit(out, minutes);
    out << ':';
    outputDoubleDigit(out, seconds);
}

/* void verifyFormattedTime(int& hours, int& minutes, int& seconds)
     Simply ensures that minutes and seconds aren't greater than 59, which would
     of course be invalid, strictly speaking. */

void verifyFormattedTime(int& hours, int& minutes, int& seconds) {
    minutes += (seconds / 60);
    seconds %= 60;

    hours += (minutes / 60);
    minutes %= 60;
}

/* newTrackList(char *listName, char *sourceFile) {
     Creates a new track list, with the name 'listName' and designated as being
     from the file 'sourceFile', and returns a pointer to it. */

TrackList* newTrackList(char *listName, char *sourceFile) {
    TrackList *result;

    result = new TrackList;
    result->first = NULL;
    result->last = NULL;
    result->name = listName;
    result->openedFrom = sourceFile;
    result->numberOfTracks = 0;

    return result;
}

/* newTrackListFromFile(char *listName, char *fileName)
     Returns a new TrackList named 'listName' and returns it, after reading in
     data from the file 'fileName'.  Returns NULL if the file cannot be read. */

TrackList* newTrackListFromFile(char *listName, char *fileName) {
    TrackList *result;
    ifstream fin;
    char *time, *artist, *title;
    int hours, minutes, seconds;
    bool isNegative, error;

    fin.open(fileName);

    if (fin.fail()) {
        return NULL; // unable to open file
    }

    result = newTrackList(listName, fileName);
    
    while(!fin.eof()) {
        artist = new char[LINE_BUFFER_SIZE];
        safeGetline(fin, artist, LINE_BUFFER_SIZE);
        removeOutsideWhitespace(&artist);

        title = new char[LINE_BUFFER_SIZE];
        safeGetline(fin, title, LINE_BUFFER_SIZE);
        removeOutsideWhitespace(&title);

        time = new char[LINE_BUFFER_SIZE];
        safeGetline(fin, time, LINE_BUFFER_SIZE);
        removeOutsideWhitespace(&time);

        if (strlen(artist) && strlen(time) && strlen(title)) {
            parseFormattedTimeFromString(time, hours, minutes, seconds,
                                         isNegative, error);
            addTrackToList(result, newTrack(title, artist,
                                          formattedTimeToSeconds(hours, minutes,
                                                  seconds, isNegative), error));
        } else {
            delete[] artist;
            delete[] title;
        }
        
        delete[] time;
    }

    fin.close();

    return result;
}

/* deleteTrackList(TrackList *target)
     Deallocates all memory used by the track list and destroys it.  'target'
     will thus be set to NULL, if it isn't already. */

void deleteTrackList(TrackList *target) {
    if (target) {
        if (target->first) {
            deleteAllTracksFromList(target);
        }

        if (target->name) {
            delete[] (target->name);
        }

        if (target->openedFrom) {
            delete[] (target->openedFrom);
        }

        delete target;
    }
}

/* Track* newTrack(int length)
     Returns a pointer to a newly allocated Track struct, setting the track
     length to the initial value as passed in 'length'. */

Track* newTrack(int length) {
    Track* temp;

    temp = new Track;
    temp->artist = NULL;
    temp->title = NULL;
    temp->lengthInSeconds = length;
    temp->warning = false;
    temp->claims = 0;
    
    return temp;
}

/* Track* newTrack(char *title)
     Returns a pointer to a newly allocated Track struct, setting the track
     name to the initial value as passed in 'name'. */

Track* newTrack(char *title) {
    Track* temp;

    temp = new Track;
    temp->title = title;
    temp->artist = NULL;
    temp->warning = false;
    temp->claims = 0;

    return temp;
}

/* Track* newTrack(char *title, char *artist, int length, bool warning)
     Returns a pointer to a newly allocated Track struct, setting the track
     name, artist and length to the initial values passed in 'name', 'artist',
     'length' and 'warn' respectively. */

Track* newTrack(char *title, char *artist, int length, bool warning) {
    Track* temp;

    temp = new Track;
    temp->title = title;
    temp->artist = artist;
    temp->lengthInSeconds = length;
    temp->warning = warning;
    temp->claims = 0;

    return temp;
}

/* deleteTrackListItem(TrackListItem *item)
     Deletes the trackListItem 'target'.  May or may not delete the track
     pointed to by the list item, depending on the list item's 'claim' property.
     'target' will always be NULL after this function. */

void deleteTrackListItem(TrackListItem *item) { 
    if (item) {
        if (item->target) {
            --(item->target->claims);

            if (item->target->claims < 1) {
                deleteTrack(item->target);
            }
        }

        delete item;
    }
}

/* deleteTrack(Track *target)
     Deletes the track 'target', meaning removes it and all its data from
     memory.  After calling this 'target' will no longer be a valid pointer and
     should not be used.  Note that this does not care what the value of the
     'claims' property is, so be very careful that you do care and check before
     you use it. */

void deleteTrack(Track *target) {
    if (target) {
        if (target->title) {
            delete[] (target->title);
        }

        if (target->artist) {
            delete[] (target->artist);
        }

        delete target;
    }
}

/* addTrackToList(TrackList *list, Track *entry)
     Adds the track 'entry' to the track list 'list', performing all the voodoo
     required to make everything work properly. */

void addTrackToList(TrackList *list, Track *entry) {
    TrackListItem *item;

    if (entry) {
        item = new TrackListItem;
        item->owner = list;
        item->target = entry;
        ++(entry->claims);
        
        if (list->last) {
            item->previous = list->last;
            list->last->next = item;
        } else {
            list->first = item;
            item->previous = NULL;
        }

        item->next = NULL;
        list->last = item;
        
        ++(list->numberOfTracks);
    }
}

/* copyTrackToList(TrackList *list, TrackListItem *entry)
     Copies the track contained in 'entry' to the track list 'list'.  The
     'entry' item will not be moved from it's present location. */

void copyTrackToList(TrackList *list, TrackListItem *entry) {
    TrackListItem *item;

    if (entry && entry->target) {
        item = new TrackListItem;
        item->owner = list;
        item->original = entry;
        item->target = entry->target;
        ++(entry->target->claims);

        if (list->last) {
            item->previous = list->last;
            list->last->next = item;
        } else {
            list->first = item;
            item->previous = NULL;
        }

        item->next = NULL;
        list->last = item;

        ++(list->numberOfTracks);
    }
}

/* removeTrackFromList(TrackListItem *item) {
     Removes the track from the list, if it is in it.  As with addTrackToList(),
     this performs all the pointer manipulations required to ensure this is a
     seamless process.  Returns true if the track was in the list and was
     successfully removed, false otherwise.
       Note that this doesn't _delete_ the track list item or track. */

bool removeTrackFromList(TrackListItem *item) {
    if (item->next) {
        if (item->previous) { // Middle of the list
            item->next->previous = item->previous;
            item->previous->next = item->next;
        } else { // First track
            item->owner->first = item->next;
            item->next->previous = NULL;
        }
    } else {
        if (item->previous) { // Last track
            item->owner->last = item->previous;
            item->previous->next = NULL;
        } else { // Not in a list
            return false;
        }
    }

    --(item->owner->numberOfTracks);

    return true;
}

/* deleteAllTracksFromList(TrackList *list)
     Removes every track from 'list', deallocating them at the same time. */

void deleteAllTracksFromList(TrackList *list) {
    TrackListItem *currentItem, *nextItem;

    currentItem = list->first;

    while (currentItem) {
        if (currentItem->next) {
            nextItem = currentItem->next;
        } else {
            nextItem = NULL;
        }

        deleteTrackListItem(currentItem);
        currentItem = nextItem;
    }

    list->numberOfTracks = 0;
}

/* findTrackByLength(TrackList *list, const int length)
     Searches the track list 'list' for the first track with a length of
     'length' seconds.  Returns a pointer to this track list item.  Returns NULL
     if no matches are found. */

TrackListItem* findTrackByLength(TrackList *list, const int length) {
    TrackListItem *currentItem;

    currentItem = list->first;

    while (currentItem) {
        if (currentItem->target->lengthInSeconds == length) {
            return currentItem;
        }

        currentItem = currentItem->next;
    }

    return NULL;
}

/* findTrackByTitle(TrackList *list, const char *name)
     Using the same process as findTrackByLength, instead compares the title of
     each track with the name provided, returning the first matching result, or
     NULL if none. */

TrackListItem* findTrackByTitle(TrackList *list, const char *name) {
    TrackListItem *currentItem;

    currentItem = list->first;

    while (currentItem) {
        if (!caseInsensitiveStringCompare(currentItem->target->title, name)) {
            return currentItem;
        }

        currentItem = currentItem->next;
    }

    return NULL;
}

/* findTrackByArtist(TrackList *list, const char *name)
     Using the same process as findTrackByLength, instead compares the artist of
     each track with the name provided, returning the first matching result, or
     NULL if none. */

TrackListItem* findTrackByArtist(TrackList *list, const char *name) {
    TrackListItem *currentItem;

    currentItem = list->first;

    while (currentItem) {
        if (!caseInsensitiveStringCompare(currentItem->target->artist, name)) {
            return currentItem;
        }

        currentItem = currentItem->next;
    }

    return NULL;
}

/* findTrackByIndex(TrackList *list, int val)
     Finds nth track, where n is 'val', in the list, starting from the first.
     Returns a pointer to said track, or NULL if the index requested is outside
     the scope of the list. */

TrackListItem* findTrackByIndex(TrackList *list, int val) {
    TrackListItem *currentItem;

    if (val < 0) {
        return NULL;
    }

    currentItem = list->first;

    while (currentItem && (val > 0)) {
        currentItem = currentItem->next;
        --val;
    }

    return currentItem;
}

/* findNextTrackByLength(const int length, TrackListItem *start)
     Continues searching from 'start' for the next track with a length of
     'length' seconds.  Returns a pointer to this track list item.  Returns NULL
     if no matches are found. */

TrackListItem* findNextTrackByLength(const int length, TrackListItem *start) {
    TrackListItem *currentItem;

    if (start) {
        currentItem = start->next;

        while (currentItem) {
            if (currentItem->target->lengthInSeconds == length) {
                return currentItem;
            }

            currentItem = currentItem->next;
        }
    }
    
    return NULL;
}

/* findNextTrackByTitle(const char *name, TrackListItem *start)
     Using the same process as findNextTrackByLength, instead compares the title
     of each track with the name provided, returning the next matching result
     after 'start', or NULL if none. */

TrackListItem* findNextTrackByTitle(const char *name, TrackListItem *start) {
    TrackListItem *currentItem;

    if (start) {
        currentItem = start->next;

        while (currentItem) {
            if (!caseInsensitiveStringCompare(currentItem->target->title,
                                              name)) {
                return currentItem;
            }

            currentItem = currentItem->next;
        }
    }
    
    return NULL;
}

/* findNextTrackByArtist(const char *name, TrackListItem *start)
     Using the same process as findNextTrackByLength, instead compares the
     artist of each track with the name provided, returning the next matching
     result after 'start', or NULL if none. */

TrackListItem* findNextTrackByArtist(const char *name, TrackListItem *start) {
    TrackListItem *currentItem;

    if (start) {
        currentItem = start->next;

        while (currentItem) {
            if (!caseInsensitiveStringCompare(currentItem->target->artist,
                                              name)) {
                return currentItem;
            }

            currentItem = currentItem->next;
        }
    }
    
    return NULL;
}

/* findAllTracksByLength(TrackList *list, const int length)
     Returns a TrackList instance containing all results for the specified
     search in the list 'list'.  Uses the functions findTrackByLength() and
     findNextTrackByLength(), so the 'length' parameter functions the same. */

TrackList* findAllTracksByLength(TrackList *list, const int length) {
    TrackList *result;
    TrackListItem *currentItem;

    result = newTrackList(NULL, NULL);
    currentItem = findTrackByLength(list, length);

    while (currentItem) {
        copyTrackToList(result, currentItem);
        currentItem = findNextTrackByLength(length, currentItem);
    }

    return result;
}

/* findAllTracksByTitle(TrackList *list, const char *name)
     Returns a TrackList instance containing all results for the specified
     search in the list 'list'.  Uses the functions findTrackByTitle() and
     findNextTrackByTitle(), so the 'name' parameter functions the same. */

TrackList* findAllTracksByTitle(TrackList *list, const char *name) {
    TrackList *result;
    TrackListItem *currentItem;

    result = newTrackList(NULL, NULL);
    currentItem = findTrackByTitle(list, name);

    while (currentItem) {
        copyTrackToList(result, currentItem);
        currentItem = findNextTrackByTitle(name, currentItem);
    }

    return result;
}

/* findAllTracksByArtist(TrackList *list, const char *name)
     Returns a TrackList instance containing all results for the specified
     search in the list 'list'.  Uses the functions findTrackByArtist() and
     findNextTrackByArtist(), so the 'name' parameter functions the same. */

TrackList* findAllTracksByArtist(TrackList *list, const char *name) {
    TrackList *result;
    TrackListItem *currentItem;

    result = newTrackList(NULL, NULL);
    currentItem = findTrackByArtist(list, name);

    while (currentItem) {
        copyTrackToList(result, currentItem);
        currentItem = findNextTrackByArtist(name, currentItem);
    }

    return result;
}

/* saveTrackListToFile(TrackList *list, const char *fileName)
     Reverse of the above; writes the song list 'list' to the file 'fileName'.
     Returns true if the save was successful, false otherwise.  Pass NULL as
     the 'fileName' parameter to use the openedFrom property of 'list'. */

bool saveTrackListToFile(TrackList *list, const char *fileName) {
    ofstream fout;
    TrackListItem *currentItem;
    Track *current;
    int hours, minutes, seconds;
    bool isNegative;

    if (fileName) {
        fout.open(fileName);
    } else {
        fout.open(list->openedFrom);
    }

    if (fout.fail()) {
        return false; // unable to open file
    }

    currentItem = list->first;

    while (currentItem) {
        current = currentItem->target;
        
        fout << current->artist << endl << current->title << endl;

        secondsToFormattedTime(current->lengthInSeconds, hours, minutes,
                               seconds, isNegative);
        outputFormattedTime(fout, hours, minutes, seconds, isNegative);
        fout << endl;

        currentItem = currentItem->next;
    }

    fout.close();

    return true;
}

/* saveDatabaseAs(TrackList *list)
     Prompts the user for a file name and saves the song list 'list' to it.
     Returns false if saving was unsuccessful *and* the user choose not to
     quit anyway.  If saving failed but they chose to quit, returns true. */

bool saveDatabaseAs(TrackList *list) {
    char *input;

    cout << "Please enter the file name to save the list \"" << list->name
         << "\".\nHit return without entering anything to save as "
         << "\"" << list->openedFrom << "\"\n?  ";

    input = new char[FILENAME_BUFFER_SIZE];
    safeGetline(cin, input, FILENAME_BUFFER_SIZE);

    if (strlen(input)) {
        if (saveTrackListToFile(list, input)) {
            cout << "\nSaved database to \"" << input << "\".  Goodbye.\n\n";
            delete[] input;
            return true;
        } else {
            cout << "\nUnable to save changes - the file \"" << input << "\" "
                 << "may be unwritable or already in use.\nDo you wish to "
                 << "quit without saving changes? (y/n)  ";
            safeGetline(cin, input, FILENAME_BUFFER_SIZE);

            if (input[0] == 'y' || input[0] == 'Y') {
                cout << "\nQuitting without saving changes.  Goodbye.\n\n";
                delete[] input;
                return true;
            } else {
                cout << "\nCancelling quit - returning to main menu.";
                delete[] input;
                return false;
            }
        }
    } else {
        if (saveTrackListToFile(list, list->openedFrom)) {
            cout << "\nSaved database to \"" << list->openedFrom << "\".\n\n";
            delete[] input;
            return true;
        } else {
            cout << "\nUnable to save changes - the file \"" << list->openedFrom
                 << "\" " << "may be unwritable or already in use.\nDo you "
                 << "wish to quit without saving changes? (y/n)  ";
            safeGetline(cin, input, FILENAME_BUFFER_SIZE);

            if (input[0] == 'y' || input[0] == 'Y') {
                cout << "\nQuitting without saving changes.  Goodbye.\n\n";
                delete[] input;
                return true;
            } else {
                cout << "\nCancelling quit - returning to main menu.";
                delete[] input;
                return false;
            }
        }
    }
}

/* copyOfString(char *data)
     Returns a copy of the string 'data'.  Useful for cloning the default track
     file name and list name's, in order to avoid malloc errors. */

char* copyOfString(char *data) {
    char *result;

    result = new char[strlen(data) + 1];
    strcpy(result, data);

    return result;
}

/* caseInsensitiveStringCompare(const char *a, const char *b)
     Compares the two strings 'a' and 'b' without regard to their case.  Returns
     -1 if 'a' is less than 'b', 1 if 'b' is less than 'a', and 0 if they are
     identical - i.e. the same as strcmp(). */

int caseInsensitiveStringCompare(const char *a, const char *b) {
    int offset, end, lengthA, lengthB, charA, charB;

    lengthA = strlen(a);
    lengthB = strlen(b);
    
    end = (lengthA < lengthB) ? lengthA : lengthB;

    for(offset = 0; offset < end; ++offset) {
        charA = toupper(a[offset]);
        charB = toupper(b[offset]);

        if (charA != charB) {
            return (charA < charB) ? -1 : 1;
        }
    }

    if (lengthA == lengthB) {
        return 0;
    } else {
        return (lengthA < lengthB) ? -1 : 1;
    }
}

/* safeGetline(istream &in, char *data, int maxLength)
     Same functionality as getline in terms of input retrieval, except that in
     the case where there is too much data it clears the error flag and buffer
     [up to the next line] so they may be used again. */

void safeGetline(istream& in, char *data, int maxLength) {
    if (cin.fail()) { // Stream already broken
        return;
    }
    
    in.getline(data, maxLength);

    if (cin.fail() && strlen(data)) {
        in.clear();
        in.ignore(10000, '\n');
    }
}
